/**
 * Copyright (c) 2003, www.pdfbox.org
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 3. Neither the name of pdfbox; nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * http://www.pdfbox.org
 *
 */
package org.pdfbox.encryption;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.io.IOException;

import java.util.Iterator;
import java.util.List;

import org.pdfbox.exceptions.CryptographyException;
import org.pdfbox.exceptions.InvalidPasswordException;

import org.pdfbox.cos.COSArray;
import org.pdfbox.cos.COSBase;
import org.pdfbox.cos.COSDictionary;
import org.pdfbox.cos.COSDocument;
import org.pdfbox.cos.COSName;
import org.pdfbox.cos.COSNumber;
import org.pdfbox.cos.COSObject;
import org.pdfbox.cos.COSStream;
import org.pdfbox.cos.COSString;

import org.pdfbox.pdmodel.PDDocument;

/**
 * This class will deal with decrypting a document.
 *
 * @author Ben Litchfield
 * @version $Revision: 1.8 $
 */
public class DecryptDocument
{
    private static final COSName ENCRYPT = COSName.getPDFName( "Encrypt" );
    private COSDocument document = null;

    private byte[] encryptionKey = null;
    private PDFEncryption encryption = new PDFEncryption();

    /**
     * Constructor.
     *
     * @param doc The document to decrypt.
     */
    public DecryptDocument( PDDocument doc )
    {
        this( doc.getDocument() );
    }

    /**
     * Constructor.
     *
     * @param doc The document to decrypt.
     */
    public DecryptDocument( COSDocument doc )
    {
        document = doc;
    }

    /**
     * This will decrypt the document.
     *
     * @param password The password for the document.
     *
     * @throws CryptographyException If there is an error decrypting the document.
     * @throws IOException If there is an error getting the stream data.
     * @throws InvalidPasswordException If the password is not a user or owner password.
     */
    public void decryptDocument( String password )
        throws CryptographyException, IOException, InvalidPasswordException
    {
        long start = System.currentTimeMillis();
        COSDictionary trailer = document.getTrailer();

        COSDictionary encryptedDictionary =
            (COSDictionary)trailer.getDictionaryObject( ENCRYPT );

        long permissions = ((COSNumber)encryptedDictionary.getDictionaryObject( COSName.getPDFName( "P" ) )).intValue();
        long revision = ((COSNumber)encryptedDictionary.getDictionaryObject( COSName.getPDFName( "R" ) )).intValue();
        COSNumber cosLength = (COSNumber)encryptedDictionary.getDictionaryObject( COSName.getPDFName( "Length" ) );
        long length = 5;
        if( cosLength != null )
        {
            length = cosLength.intValue() / 8;
        }
        COSString id = (COSString)document.getDocumentID().get( 0 );
        COSString userPassword = (COSString)encryptedDictionary.getDictionaryObject( COSName.getPDFName( "U" ) );
        COSString ownerPassword = (COSString)encryptedDictionary.getDictionaryObject( COSName.getPDFName( "O" ) );

        boolean isUserPassword =
            encryption.isUserPassword( password.getBytes(), userPassword.getBytes(),
                ownerPassword.getBytes(),
                permissions, id.getBytes(), revision, length );
        boolean isOwnerPassword =
            encryption.isOwnerPassword( password.getBytes(), userPassword.getBytes(),
                ownerPassword.getBytes(), permissions, id.getBytes(), revision, length );

        if( isUserPassword )
        {
            encryptionKey =
                encryption.computeEncryptedKey(
                    password.getBytes(), ownerPassword.getBytes(),
                    permissions, id.getBytes(), revision, length );
        }
        else if( isOwnerPassword )
        {
            byte[] computedUserPassword =
                encryption.getUserPassword(
                    password.getBytes(),
                    ownerPassword.getBytes(),
                    revision,
                    length );
            encryptionKey =
                encryption.computeEncryptedKey(
                    computedUserPassword, ownerPassword.getBytes(),
                    permissions, id.getBytes(), revision, length );
        }
        else
        {
            throw new InvalidPasswordException( "Error: The supplied password does not match " +
                                                "either the owner or user password in the document." );
        }

        List allObjects = document.getObjects();
        Iterator objectIter = allObjects.iterator();
        while( objectIter.hasNext() )
        {
            decryptObject( (COSObject)objectIter.next() );
        }
        document.setOriginalEncryptionDictionary( encryptedDictionary );
        trailer.removeItem( ENCRYPT );
        long stop = System.currentTimeMillis();
    }

    /**
     * This will print a byte array as a hex string to standard output.
     *
     * @param data The array to print.
     */
    private static void printHexString( byte[] data )
    {
        for( int i=0; i<data.length; i++ )
        {
            int nextByte = (data[i] + 256)%256;
            String hexString = Integer.toHexString( nextByte );
            if( hexString.length() < 2 )
            {
                hexString = "0" + hexString;
            }
            System.out.print( hexString );
            if( i != 0 && (i+1) % 2 == 0 )
            {
                System.out.print( " " );
            }
            else if( i!= 0 &&i % 20 == 0 )
            {
                System.out.println();
            }
        }
        System.out.println();
    }

    /**
     * This will decrypt an object in the document.
     *
     * @param object The object to decrypt.
     *
     * @throws CryptographyException If there is an error decrypting the stream.
     * @throws IOException If there is an error getting the stream data.
     */
    private void decryptObject( COSObject object )
        throws CryptographyException, IOException
    {
        long objNum = object.getObjectNumber().intValue();
        long genNum = object.getGenerationNumber().intValue();
        COSBase base = object.getObject();
        decrypt( base, objNum, genNum );
    }

    /**
     * This will dispatch to the correct method.
     *
     * @param obj The object to decrypt.
     * @param objNum The object number.
     * @param genNum The object generation Number.
     *
     * @throws CryptographyException If there is an error decrypting the stream.
     * @throws IOException If there is an error getting the stream data.
     */
    private void decrypt( Object obj, long objNum, long genNum )
        throws CryptographyException, IOException
    {
        if( obj instanceof COSString )
        {
            decryptString( (COSString)obj, objNum, genNum );
        }
        else if( obj instanceof COSDictionary )
        {
            decryptDictionary( (COSDictionary)obj, objNum, genNum );
        }
        else if( obj instanceof COSStream )
        {
            decryptStream( (COSStream)obj, objNum, genNum );
        }
        else if( obj instanceof COSArray )
        {
            decryptArray( (COSArray)obj, objNum, genNum );
        }
    }

    /**
     * This will decrypt a stream.
     *
     * @param stream The stream to decrypt.
     * @param objNum The object number.
     * @param genNum The object generation number.
     *
     * @throws CryptographyException If there is an error getting the stream.
     * @throws IOException If there is an error getting the stream data.
     */
    private void decryptStream( COSStream stream, long objNum, long genNum )
        throws CryptographyException, IOException
    {
        decryptDictionary( stream.getDictionary(), objNum, genNum );
        InputStream encryptedStream = stream.getFilteredStream();
        encryption.encryptData( objNum,
                                genNum,
                                encryptionKey,
                                encryptedStream,
                                stream.createFilteredStream() );
    }

    /**
     * This will decrypt a dictionary.
     *
     * @param dictionary The dictionary to decrypt.
     * @param objNum The object number.
     * @param genNum The object generation number.
     *
     * @throws CryptographyException If there is an error decrypting the document.
     * @throws IOException If there is an error creating a new string.
     */
    private void decryptDictionary( COSDictionary dictionary, long objNum, long genNum )
        throws CryptographyException, IOException
    {
        Iterator values = dictionary.getValues().iterator();
        while( values.hasNext() )
        {
            decrypt( values.next(), objNum, genNum );
        }
    }

    /**
     * This will decrypt a string.
     *
     * @param string the string to decrypt.
     * @param objNum The object number.
     * @param genNum The object generation number.
     *
     * @throws CryptographyException If an error occurs during decryption.
     * @throws IOException If an error occurs writing the new string.
     */
    private void decryptString( COSString string, long objNum, long genNum )
        throws CryptographyException, IOException
    {
        ByteArrayInputStream data = new ByteArrayInputStream( string.getBytes() );
        ByteArrayOutputStream buffer = new ByteArrayOutputStream();
        encryption.encryptData( objNum,
                                genNum,
                                encryptionKey,
                                data,
                                buffer );
        string.reset();
        string.append( buffer.toByteArray() );
    }

    /**
     * This will decrypt an array.
     *
     * @param array The array to decrypt.
     * @param objNum The object number.
     * @param genNum The object generation number.
     *
     * @throws CryptographyException If an error occurs during decryption.
     * @throws IOException If there is an error accessing the data.
     */
    private void decryptArray( COSArray array, long objNum, long genNum )
        throws CryptographyException, IOException
    {
        for( int i=0; i<array.size(); i++ )
        {
            decrypt( array.get( i ), objNum, genNum );
        }
    }
}