/**
 * Copyright (c) 2003, www.pdfbox.org
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 3. Neither the name of pdfbox; nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * http://www.pdfbox.org
 *
 */
package org.pdfbox.examples.fdf;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;

import org.pdfbox.cos.COSArray;
import org.pdfbox.cos.COSDictionary;
import org.pdfbox.cos.COSDocument;
import org.pdfbox.cos.COSName;
import org.pdfbox.cos.COSString;

import org.pdfbox.pdfparser.PDFParser;

import org.pdfbox.pdmodel.PDDocument;
import org.pdfbox.pdmodel.PDDocumentCatalog;

import org.pdfbox.pdmodel.annotation.interactive.PDAcroForm;
import org.pdfbox.pdmodel.annotation.interactive.PDField;

import org.pdfbox.pdfwriter.COSWriter;

import org.pdfbox.examples.AbstractExample;

/**
 * This example will take a PDF document and fill the fields with data from the
 * FDF fields.  This will not set the appearance stream in the document.  If you
 * open a document with fields that have been set by this application then you
 * need to click on the field to get it to show on the screen.
 *
 * I would be interested in anybody that is willing to submit some code that shows
 * how to create the appearance stream.
 *
 * @author  Ben Litchfield (ben@csh.rit.edu)
 * @version $Revision: 1.7 $
 */
public class ImportFDF extends AbstractExample
{
    /**
     * Creates a new instance of ImportFDF
     */
    public ImportFDF()
    {
    }

    /**
     * This will takes the values from the fdf document and import them into the
     * PDF document.
     *
     * @param pdfDocument The document to put the fdf data into.
     * @param fdfDocument The FDF document to get the data from.
     *
     * @throws IOException If there is an error setting the data in the field.
     */
    public void importFDF( PDDocument pdfDocument, COSDocument fdfDocument ) throws IOException
    {
        PDDocumentCatalog docCatalog = pdfDocument.getDocumentCatalog();
        PDAcroForm acroForm = docCatalog.getAcroForm();
        acroForm.setCacheFields( true );

        COSDictionary trailer = fdfDocument.getTrailer();
        COSDictionary root = (COSDictionary)trailer.getDictionaryObject( COSName.ROOT );
        COSDictionary fdf = (COSDictionary)root.getDictionaryObject( COSName.getPDFName( "FDF" ) );
        if( fdf != null )
        {
            COSArray fields = (COSArray)fdf.getDictionaryObject( COSName.getPDFName( "Fields" ) );
            for( int i=0; i<fields.size(); i++ )
            {
                COSDictionary fdfField = (COSDictionary)fields.getObject( i );
                COSString fieldName = (COSString)fdfField.getDictionaryObject( COSName.getPDFName( "T" ) );
                System.out.println( "Processing " + fieldName );
                PDField docField = acroForm.getField( fieldName.getString() );
                if( docField != null )
                {
                    docField.importFDF( fdfField );
                }
            }
        }
    }

    /**
     * This will import an fdf document and write out another pdf.
     * <br />
     * see usage() for commandline
     *
     * @param args command line arguments
     *
     * @throws Exception If there is an error importing the FDF document.
     */
    public static void main(String[] args) throws Exception
    {
        PDDocument pdf = null;
        COSDocument fdf = null;
        FileInputStream pdfStream = null;
        FileInputStream fdfStream = null;
        FileOutputStream output = null;
        PDFParser pdfParser = null;
        PDFParser fdfParser = null;
        COSWriter writer = null;

        try
        {
            if( args.length != 3 )
            {
                usage();
            }
            else
            {
                ImportFDF importer = new ImportFDF();
                pdfStream = new FileInputStream( args[0] );
                fdfStream = new FileInputStream( args[1] );
                output = new FileOutputStream( args[2] );
                pdfParser = new PDFParser( pdfStream );
                fdfParser = new PDFParser( fdfStream );
                pdfParser.parse();
                fdfParser.parse();

                pdf = pdfParser.getPDDocument();
                fdf = fdfParser.getDocument();
                importer.importFDF( pdf, fdf );

                writer = new COSWriter( output );
                writer.write( pdf );
            }
        }
        finally
        {
            close( fdf );
            close( pdf );
            close( pdfStream );
            close( fdfStream );
            close( output );
            close( writer );
        }
    }
    /**
     * This will print out a message telling how to use this example.
     */
    private static void usage()
    {
        System.err.println( "usage: org.pdfbox.examples.fdf.ImportFDF <pdf-file> <fdf-file> <output-file>" );
    }
}