/**
 * Copyright (c) 2003, www.pdfbox.org
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 3. Neither the name of pdfbox; nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * http://www.pdfbox.org
 *
 */
package org.pdfbox.pdmodel;

import org.pdfbox.cos.COSDictionary;
import org.pdfbox.cos.COSFloat;
import org.pdfbox.cos.COSInteger;
import org.pdfbox.cos.COSName;
import org.pdfbox.cos.COSNumber;

/**
 * This represents the Graphics State Parameter Dictionary.
 *
 * @author  Ben Litchfield (ben@csh.rit.edu)
 * @version $Revision: 1.2 $
 */
public class PDExtGState
{
    private static final COSName LW = COSName.getPDFName( "LW" );
    private static final COSName LC = COSName.getPDFName( "LC" );
    private static final COSName LJ = COSName.getPDFName( "LJ" );

    private COSDictionary resources;

    /**
     * This will create a new PDExtGState using the COSDictionary to back to.
     *
     * @param state The state of the resources dictionary.
     */
    public PDExtGState( COSDictionary state )
    {
        resources = state;
    }

    /**
     * Creates a new instance of PDExtGState.
     */
    public PDExtGState()
    {
        resources = new COSDictionary();
        resources.setItem( COSName.TYPE, COSName.getPDFName( "ExtGState" ) );
    }

    /**
     * This will get the line width.  This will return null if there is no line width
     *
     * @return null or the LW value of the dictionary.
     */
    public Float getLineWidth()
    {
        return getFloatItem( LW );
    }

    /**
     * This will set the line width.
     *
     * @param width The line width for the object.
     */
    public void setLineWidth( Float width )
    {
        setFloatItem( LW, width );
    }

    /**
     * This will get the line cap style
     *
     * @return null or the LC value of the dictionary.
     */
    public Long getLineCapStyle()
    {
        return getLongItem( LC );
    }

    /**
     * This will set the line cap style for the graphics state.
     *
     * @param style The new line cap style to set.
     */
    public void setLineCapStyle( Long style )
    {
        setLongItem( LC, style );
    }

    /**
     * This will get the line join style.
     *
     * @return null or the LJ value in the dictionary.
     */
    public Long getLineJoinStyle()
    {
        return getLongItem( LJ );
    }

    /**
     * This will set the line join style.
     *
     * @param style The new line join style.
     */
    public void setLineJoinStyle( Long style )
    {
        setLongItem( LJ, style );
    }

    /**
     * This will get a float item from the dictionary.
     *
     * @param key The key to the item.
     *
     * @return The value for that item.
     */
    private Float getFloatItem( COSName key )
    {
        Float retval = null;
        COSNumber value = (COSNumber)resources.getDictionaryObject( key );
        if( value != null )
        {
            retval = new Float( value.floatValue() );
        }
        return retval;
    }

    /**
     * This will set a float object.
     *
     * @param key The key to the data that we are setting.
     * @param value The value that we are setting.
     */
    private void setFloatItem( COSName key, Float value )
    {
        if( value == null )
        {
            resources.removeItem( key );
        }
        else
        {
            resources.setItem( key, new COSFloat( value.floatValue() ) );
        }
    }

    /**
     * This will get an int item from the dictionary.
     *
     * @param key The key to the item.
     *
     * @return The value for that item.
     */
    private Long getLongItem( COSName key )
    {
        Long retval = null;
        COSNumber value = (COSNumber)resources.getDictionaryObject( key );
        if( value != null )
        {
            retval = new Long( value.intValue() );
        }
        return retval;
    }

    /**
     * This will set an integer object.
     *
     * @param key The key to the data that we are setting.
     * @param value The value that we are setting.
     */
    private void setLongItem( COSName key, Long value )
    {
        if( value == null )
        {
            resources.removeItem( key );
        }
        else
        {
            resources.setItem( key, new COSInteger( value.longValue() ) );
        }
    }
}