/**
 * Copyright (c) 2003, www.pdfbox.org
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 3. Neither the name of pdfbox; nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * http://www.pdfbox.org
 *
 */
package org.pdfbox.pdmodel.annotation.interactive;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import org.pdfbox.cos.COSArray;
import org.pdfbox.cos.COSBase;
import org.pdfbox.cos.COSDictionary;
import org.pdfbox.cos.COSFloat;
import org.pdfbox.cos.COSInteger;
import org.pdfbox.cos.COSName;
import org.pdfbox.cos.COSStream;
import org.pdfbox.cos.COSString;

/**
 * This one took me a while, but i'm proud to say that it handles
 * the appearance of a textbox. This allows you to apply a value to
 * a field in the document and handle the appearance so that the
 * value is actually visible too.
 * The problem was described by Ben Litchfield, the author of the
 * example: org.pdfbox.examlpes.fdf.ImportFDF. So Ben, here is the
 * solution.
 *
 * @author sug
 * @version $Revision: 1.1 $
 */
public class PDAppearance
{
    private PDTextbox parent;

    private COSStream ap;
    private COSString da;
    private COSArray bbox;

    private float boxX;
    private float boxY;
    private float boxWidth;
    private float boxHeight;

    private float lineWidth;

    private String stream;

    private String value;

    private String font;
    private float fontSize;

    private PDAcroForm acroForm;


    /**
     * Constructs a COSAppearnce from the given field
     *
     * @param theAcroForm the acro form that this field is part of.
     * @param field the field which you wish to control the appearance of
     */
    public PDAppearance( PDAcroForm theAcroForm, PDTextbox field )
    {
        acroForm = theAcroForm;
        parent = field;
        COSDictionary a = (COSDictionary) field.getDictionary().getDictionaryObject(COSName.getPDFName("AP"));
        this.ap = (COSStream) a.getDictionaryObject(COSName.getPDFName("N"));
        this.da = getDefaultAppearance();

        bbox = (COSArray) ap.getDictionaryObject(COSName.getPDFName("BBox"));

        boxX = floatValue(bbox.get(0));
        boxY = floatValue(bbox.get(1));
        boxWidth = floatValue(bbox.get(2));
        boxHeight = floatValue(bbox.get(3));

        stream = getOriginalStream();
        lineWidth = getLineWidth();
    }

    /**
     * Returns the default apperance of a textbox. If the textbox
     * does not have one, then it will be taken from the AcroForm.
     * @return The DA element
     */
    private COSString getDefaultAppearance()
    {
        COSString dap = parent.getDefaultAppearance();
        if (dap == null)
        {
            dap = (COSString) acroForm.getDictionary().getDictionaryObject(COSName.getPDFName("DA"));
        }
        return dap;
    }

    /**
     * Takes the fontname and finds the reference to it
     * @param fontname the name of the font. e.g. "Helv"
     * @return a COSBase the reference to the font element.
     */
    private COSBase getFontReference(String fontname)
    {
        COSDictionary dr = (COSDictionary)acroForm.getDictionary().getDictionaryObject(COSName.getPDFName("DR"));
        COSDictionary font = (COSDictionary) dr.getDictionaryObject(COSName.FONT);
        return font.getDictionaryObject(COSName.getPDFName(fontname));
    }

    /**
     * returns the float value of either a COSInteger or COSFloat
     * @param item the item to get the floatValue from
     * @return the float value of item
     */
    private static float floatValue(COSBase item)
    {
        float r = 0f;
        if (item instanceof COSInteger)
        {
            r = ((COSInteger) item).floatValue();
        }
        else if (item instanceof COSFloat)
        {
            r = ((COSFloat) item).floatValue();
        }
        return r;
    }

    /**
     * extracts the original appearance stream into a string
     * @return the original appearance stream
     */
    private String getOriginalStream()
    {
        byte[] b = null;
        try
        {
            InputStream is = ap.getUnfilteredStream();
            b = new byte[is.available()];
            is.read(b);
        }
        catch (IOException e)
        {
            e.printStackTrace();
        }
        return new String(b);
    }

    /**
     * Tests if the apperance stream already contains content
     * @return true if it contains any content
     */
    public boolean containsContent()
    {
        return (stream.indexOf("BMC") > -1);
    }

    /**
     * This is the public method for setting the appearance stream.
     *
     * @param value the String value which the apperance shoud represent
     *
     * @throws IOException If there is an error creating the stream.
     */
    public void setAppearanceValue(String value) throws IOException
    {
        this.value = value;
        if (!containsContent())
        {
            StringBuffer sb = new StringBuffer(stream);
            sb.append(" /Tx BMC q ");
            sb.append(getInnerX()+" ").append(getInnerY()+" ");
            sb.append(getAvailableWidth()+" ").append(getAvailableHeight()+" ");
            sb.append("re W n ");
            sb.append(getTextColor());
            sb.append(" BT\n");
            sb.append(getFont()+"\n");
            sb.append(getTextPosition()+"\n");
            sb.append("0 TL\n");
            sb.append("(" + value + ") Tj\n");
            sb.append("ET Q EMC");
            stream = sb.toString();
            writeToStream();
            setFontReference();
        }
        else
        {
            int i1 = stream.indexOf("(");
            int i2 = stream.indexOf(")");
            if (i1 > -1 && i2 > -1)
            {
                stream =
                    stream.substring(0, i1 + 1)
                        + value
                        + stream.substring(i2);
                writeToStream();
            }
        }
    }

    /**
     * Writes the stream to the actual stream in the COSStream.
     *
     * @throws IOException If there is an error writing to the stream
     */
    private void writeToStream() throws IOException
    {
        OutputStream out = ap.createUnfilteredStream();
        out.write(stream.getBytes());
    }

    /**
     * This method should only be called if the appearance stream have not ben set before.
     * It sets the reference to the font dictionary
     *
     */
    private void setFontReference()
    {
        COSDictionary resources = (COSDictionary)ap.getDictionaryObject(COSName.RESOURCES);
        COSArray ar = (COSArray) resources.getDictionaryObject(COSName.getPDFName("ProcSet"));
        ar.add(COSName.getPDFName("Text"));
        COSDictionary fontDescriptor = new COSDictionary();
        String fontname = getFontName();
        COSBase fontref = getFontReference(fontname);
        String a = null;
        fontDescriptor.setItem(COSName.getPDFName(fontname), fontref);
        resources.setItem(COSName.FONT, fontDescriptor);
    }

    /**
     * w in an appearance stream represents the lineWidth
     * @return the linewidth
     */
    private float getLineWidth()
    {
        int w = stream.indexOf("w");
        String fl = "";
        for(int i = w-2; i > 0 &&
                         i < stream.length() &&
                         (Character.isDigit(stream.charAt(i)) || stream.charAt(i) == '.'); i--)
        {
            fl = stream.charAt(i) + fl;
        }
        float retval = 0;
        try
        {
            retval = Float.parseFloat(fl);
        }
        catch( NumberFormatException e )
        {
            //default to zero for line width.
        }
        return retval;
    }

    /**
     * returns the x position of the start of the box.
     * @return the x position of the start of the box
     */
    private float getInnerX()
    {
        return boxX + lineWidth;
    }

    /**
     * returns the x position of the start of the box.
     * @return the y position of the start of the box
     */
    private float getInnerY()
    {
        return boxY + lineWidth;
    }

    /**
     * calculates the available width of the box
     * @return the calculated available width of the box
     */
    private float getAvailableWidth()
    {
        return boxWidth - 2 * lineWidth;
    }

    /**
     * calculates the available height of the box
     * @return the calculated available height of the box
     */
    private float getAvailableHeight()
    {
        return boxHeight - 2 * lineWidth;
    }

    /**
     * extracts the textcolor from the da value
     * @return the text color
     */
    private String getTextColor()
    {
        if (da != null)
        {
            String color = da.getString();
            color = color.substring(color.indexOf("Tf")+2);
            return color;
        }
        else
        {
            return "0 g"; // "0 g" = black
        }
    }

    /**
     * returns the complete pdf string for representing a font
     * @return the font description
     */
    private String getFont()
    {
        if (font == null)
        {
            if (da != null)
            {
                font = da.getString();
                font = font.substring(0, font.indexOf("Tf")+2);
                String fontsize = font.substring(font.indexOf(" ")+1, font.lastIndexOf(" "));
                int indexOfZero = font.indexOf( '0' );
                if( fontsize.equals("0") && indexOfZero >= 0 )
                {
                    font = font.substring( 0, indexOfZero )
                        + Float.toString(calculateFontSize())
                        + font.substring( indexOfZero+1, font.length() );
                    //This is only avail in JDK 1.4, and we are going to
                    //try to keep 1.3 compatible for a little while longer.
                    //font = font.replaceFirst("0", Float.toString(calculateFontSize()));
                }
            }
            else
            {
                font = "/Helv " + calculateFontSize() + " Tf";
            }
        }
        return font;
    }

    /**
     * extracts the font name from the da value
     * @return the font name
     */
    private String getFontName()
    {
        String fontname = null;
        if (da != null)
        {
            fontname = da.getString();
            fontname = fontname.substring(
                fontname.indexOf("/")+1,
                fontname.indexOf(" "));
        }
        else
        {
            fontname = "Helv";
        }
        return fontname;
    }

    /**
     * extracts the fontsize from the da if != null otherwise
     * it calculates the fontsize.
     * @return the fontsize
     */
    private float getFontSize()
    {
        if (fontSize == 0.0f)
        {
            if (da == null)
            {
                fontSize = calculateFontSize();
            }
            else
            {
                if (font == null)
                {
                    getFont();
                }
                int s1 = font.indexOf(" ");
                int s2 = font.indexOf(" ", s1+1);
                String s = font.substring(s1,s2);
                fontSize = Float.parseFloat(s);
            }
        }
        return fontSize;
    }

    /**
     * My "not so great" method for calculating the fontsize.
     * It does not work superb, but it handles ok.
     * @return the calculated font-size
     */
    private float calculateFontSize()
    {
        float fsByHeight = getAvailableHeight()-2*lineWidth;
        if (value != null)
        {
            if (value.length() == 0)
            {
                return fsByHeight;
            }
            float length = Float.parseFloat(Integer.toString(value.length()));
            float fs = getAvailableWidth() / length * 2;
            if (fs > fsByHeight)
            {
                fs = fsByHeight;
            }
            return fs;
        }
        else
        {
            return fsByHeight;
        }
    }

    /**
     * Calculates where to start putting the text in the box.
     * The positioning is not quite as accurate as when Acrobat
     * places the elements, but it works though.
     * @return the sting for representing the start position of the text
     */
    private String getTextPosition()
    {
        if (fontSize == 0.0f)
        {
            getFontSize();
        }
        float pos = 0.0f;
        if (parent.isMultiline())
        {
            int rows = (int) (getAvailableHeight() / ((int) fontSize));
            pos = ((rows)*fontSize)-fontSize;
        }
        else
        {
            pos = fontSize*(4f / 10f);
            pos = boxHeight-fontSize-getLineWidth();
        }
        return "2 "+ pos + " Td";
    }
}