/**
 * Copyright (c) 2003, www.pdfbox.org
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 3. Neither the name of pdfbox; nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * http://www.pdfbox.org
 *
 */
package org.pdfbox.pdmodel.annotation.interactive;

import java.util.Vector;

import org.pdfbox.cos.COSArray;
import org.pdfbox.cos.COSDictionary;
import org.pdfbox.cos.COSName;

/**
 * A class for handling the PDF field as a Radio Collection.
 * This class automatically keeps track of the child radio buttons
 * in the collection.
 *
 * @see org.pdfbox.pdmodel.field.PDRadioButton
 * @author sug
 * @version $Revision: 1.1 $
 */
public class PDRadioCollection extends PDField
{
    private COSArray kids;
    private Vector subButtons;

    /**
     * @see org.pdfbox.pdmodel.field.PDField#PDField(org.pdfbox.cos.COSDictionary)
     *
     * @param field
     */
    public PDRadioCollection(COSDictionary field)
    {
        super(field);
        parseSubButtons();
    }

    /**
     * This setValue method iterates the collection of radiobuttons
     * and checks or unchecks each radiobutton according to the
     * given value.
     * If the value is not represented by any of the radiobuttons,
     * then none will be checked.
     *
     * @see org.pdfbox.pdmodel.field.PDField#setValue(java.lang.String)
     */
    public void setValue(String value)
    {
        COSName val = COSName.getPDFName(value);
        for (int i = 0; i < subButtons.size(); i++)
        {
            PDRadioButton btn = (PDRadioButton) subButtons.get(i);
            if (btn.representsValue(val))
            {
                btn.check();
            }
            else
            {
                btn.unCheck();
            }
        }
    }

    /**
     * Iterates each of the child radiobuttons into a Vector for
     * easy access.
     */
    private void parseSubButtons()
    {
        kids = (COSArray) getDictionary().getDictionaryObject(COSName.KIDS);
        subButtons = new Vector();
        for (int i = 0; i < kids.size(); i++)
        {
            PDRadioButton crb = new PDRadioButton((COSDictionary) kids.getObject(i));
            subButtons.add(crb);
        }
    }
}