/**
 * Copyright (c) 2003, www.pdfbox.org
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 3. Neither the name of pdfbox; nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * http://www.pdfbox.org
 *
 */
package org.pdfbox.pdmodel.annotation.interactive;

import org.pdfbox.cos.COSArray;
import org.pdfbox.cos.COSDictionary;
import org.pdfbox.cos.COSInteger;
import org.pdfbox.cos.COSName;
import org.pdfbox.cos.COSString;

import java.io.IOException;

/**
 * A class for handling the PDF field as a textbox.
 *
 * @author sug
 * @version $Revision: 1.2 $
 */
public class PDTextbox extends PDChoiceField
{
    /**
     * Ff    A set of flags specifying various characteristics of the field.
     */
    private COSInteger ff;
    /**
     * AP    An apearance dictionary specifying the appearance
     */
    private COSDictionary ap;
    /**
     * DA    Default appearance
     */
    private COSString da;

    private PDAppearance appearance;

    private PDAcroForm acroForm;

    /**
     * @see org.pdfbox.pdmodel.field.PDField#COSField(org.pdfbox.cos.COSDictionary)
     *
     * @param theAcroForm The acroForm for this field.
     * @param field The field's dictionary.
     */
    public PDTextbox( PDAcroForm theAcroForm, COSDictionary field)
    {
        super(field);
        acroForm = theAcroForm;

        // Get the Ff
        ff = (COSInteger) field.getDictionaryObject( COSName.getPDFName( "Ff" ) );

        // Get the AP
        COSArray kids = (COSArray) field.getDictionaryObject(COSName.KIDS);
        if (kids != null)
        {
            COSDictionary cd = (COSDictionary) kids.getObject(0);
            ap = (COSDictionary) cd.getDictionaryObject(COSName.getPDFName("AP"));
        }
        else
        {
           ap = (COSDictionary) field.getDictionaryObject(COSName.getPDFName("AP"));
        }
        da = (COSString) field.getDictionaryObject(COSName.getPDFName("DA"));
    }

    /**
     * @see org.pdfbox.pdmodel.field.PDField#setValue(java.lang.String)
     *
     * @param value The new value for this text field.
     *
     * @throws IOException If there is an error calculating the appearance stream.
     */
    public void setValue(String value) throws IOException
    {
        super.setValue(value);
        if( getDictionary().getDictionaryObject( COSName.getPDFName( "AP" ) ) != null )
        {
            if (appearance == null)
            {
                this.appearance = new PDAppearance( acroForm, this );
            }
            appearance.setAppearanceValue(value);
        }
    }

    /**
     * @see org.pdfbox.pdmodel.field.PDField#setReadonly()
     */
    public void setReadonly()
    {
        setFfFlag(1,true);
    }

    /**
     *
     * @return true if the field is marked as readonly
     */
    public boolean isReadonly()
    {
        return getFfFlag(1);
    }

    /**
     * @return true if the field is multiline
     */
    public boolean isMultiline()
    {
        boolean ml = getFfFlag(13);
        return ml;
    }

    /**
     * @return the DA element of the dictionary object
     */
    protected COSString getDefaultAppearance()
    {
        return da;
    }

}