/**
 * Copyright (c) 2003, www.pdfbox.org
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 3. Neither the name of pdfbox; nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * http://www.pdfbox.org
 *
 */
package org.pdfbox.pdmodel.common;

import org.pdfbox.cos.COSArray;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;

/**
 * This is an implementation of a List that will sync its contents to a COSArray.
 *
 * @author  Ben Litchfield (ben@csh.rit.edu)
 * @version $Revision: 1.1 $
 */
public class COSArrayList implements List
{
    private COSArray array;
    private List actual;

    /**
     * Constructor
     *
     * @param actualList The list of standard java objects
     * @param cosArray The COS array object to sync to.
     */
    public COSArrayList( List actualList, COSArray cosArray )
    {
        actual = actualList;
        array = cosArray;
    }

    /**
     * @see List#size()
     */
    public int size()
    {
        return actual.size();
    }

    /**
     * @see List#size()
     */
    public boolean isEmpty()
    {
        return actual.isEmpty();
    }

    /**
     * @see List#contains()
     */
    public boolean contains(Object o)
    {
        return actual.contains(o);
    }

    /**
     * @see List#iterator()
     */
    public Iterator iterator()
    {
        return actual.iterator();
    }

    /**
     * @see List#toArray()
     */
    public Object[] toArray()
    {
        return actual.toArray();
    }

    /**
     * @see List#toArray()
     */
    public Object[] toArray(Object[] a)
    {
        return actual.toArray(a);

    }

    /**
     * @see List#add()
     */
    public boolean add(Object o)
    {
        array.add( ((COSObjectable)o).getCOSObject() );
        return actual.add(o);
    }

    /**
     * @see List#remove()
     */
    public boolean remove(Object o)
    {
        array.remove( ((COSObjectable)o).getCOSObject() );
        return actual.remove( o );
    }

    /**
     * @see List#containsAll()
     */
    public boolean containsAll(Collection c)
    {
        return actual.containsAll( c );
    }

    /**
     * @see List#addAll()
     */
    public boolean addAll(Collection c)
    {
        array.addAll( toCOSObjectList( c ) );
        return actual.addAll( c );
    }

    /**
     * @see List#addAll()
     */
    public boolean addAll(int index, Collection c)
    {
        array.addAll( index, toCOSObjectList( c ) );
        return actual.addAll( index, c );
    }

    private List toCOSObjectList( Collection list )
    {
        List cosObjects = new ArrayList();
        Iterator iter = list.iterator();
        while( iter.hasNext() )
        {
            COSObjectable object = (COSObjectable)iter.next();
            cosObjects.add( object );
        }
        return cosObjects;
    }

    /**
     * @see List#removeAll()
     */
    public boolean removeAll(Collection c)
    {
        array.removeAll( toCOSObjectList( c ) );
        return actual.removeAll( c );
    }

    /**
     * @see List#retainAll()
     */
    public boolean retainAll(Collection c)
    {
        array.retainAll( toCOSObjectList( c ) );
        return actual.retainAll( c );
    }

    /**
     * @see List#clear()
     */
    public void clear()
    {
        actual.clear();
        array.clear();
    }

    /**
     * @see List#equals()
     */
    public boolean equals(Object o)
    {
        return actual.equals( o );
    }

    /**
     * @see List#hashCode()
     */
    public int hashCode()
    {
        return actual.hashCode();
    }

    /**
     * @see List#get()
     */
    public Object get(int index)
    {
        return actual.get( index );

    }

    /**
     * @see List#set()
     */
    public Object set(int index, Object element)
    {
        array.set( index, ((COSObjectable)element).getCOSObject() );
        return actual.set( index, element );
    }

    /**
     * @see List#add()
     */
    public void add(int index, Object element)
    {
        actual.add( index, element );
        array.add( index, ((COSObjectable)element).getCOSObject() );

    }

    /**
     * @see List#remove()
     */
    public Object remove(int index)
    {
        array.remove( index );
        return actual.remove( index );
    }

    /**
     * @see List#indexOf()
     */
    public int indexOf(Object o)
    {
        return actual.indexOf( o );
    }

    /**
     * @see List#lastIndexOf()
     */
    public int lastIndexOf(Object o)
    {
        return actual.indexOf( o );

    }

    /**
     * @see List#listIterator()
     */
    public ListIterator listIterator()
    {
        return actual.listIterator();
    }

    /**
     * @see List#listIterator()
     */
    public ListIterator listIterator(int index)
    {
        return actual.listIterator( index );
    }

    /**
     * @see List#subList()
     */
    public List subList(int fromIndex, int toIndex)
    {
        return actual.subList( fromIndex, toIndex );
    }
}