/**
 * Copyright (c) 2003, www.pdfbox.org
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 3. Neither the name of pdfbox; nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * http://www.pdfbox.org
 *
 */
package org.pdfbox.pdmodel.font;

import org.pdfbox.afmparser.AFMParser;

import org.pdfbox.afmtypes.FontMetric;

import org.pdfbox.cmapparser.CMapParser;

import org.pdfbox.cmaptypes.CMap;

import org.pdfbox.encoding.AFMEncoding;
import org.pdfbox.encoding.DictionaryEncoding;
import org.pdfbox.encoding.Encoding;
import org.pdfbox.encoding.EncodingManager;

import org.pdfbox.cos.COSArray;
import org.pdfbox.cos.COSBase;
import org.pdfbox.cos.COSDictionary;
import org.pdfbox.cos.COSInteger;
import org.pdfbox.cos.COSName;
import org.pdfbox.cos.COSNumber;
import org.pdfbox.cos.COSStream;


import org.pdfbox.pdmodel.common.COSObjectable;

import org.apache.log4j.Category;

import java.io.InputStream;
import java.io.IOException;

import java.util.HashMap;
import java.util.Map;

/**
 * This is the base class for all PDF fonts.
 *
 * @author Ben Litchfield (ben@csh.rit.edu)
 * @version $Revision: 1.2 $
 */
public abstract class PDFont implements COSObjectable
{
    private static Category log = Category.getInstance(PDFont.class.getName());

    /**
     * The cos dictionary for this font.
     */
    protected COSDictionary font;

    /**
     * This is only used if this is a font object and it has an encoding.
     */
    private Encoding fontEncoding = null;
    /**
     * This is only used if this is a font object and it has an encoding and it is
     * a type0 font with a cmap.
     */
    private CMap cmap = null;

    private static Map afmResources = null;
    private static Map afmObjects = null;

    static
    {
        afmResources = new HashMap();
        afmObjects = new HashMap();

        afmResources.put( COSName.getPDFName( "Courier-Bold" ), "Resources/afm/COB_____.AFM" );
        afmResources.put( COSName.getPDFName( "Courier-BoldOblique" ), "Resources/afm/COBO____.AFM" );
        afmResources.put( COSName.getPDFName( "Courier" ), "Resources/afm/COM_____.AFM" );
        afmResources.put( COSName.getPDFName( "Courier-Oblique" ), "Resources/afm/COO_____.AFM" );
        afmResources.put( COSName.getPDFName( "Helvetica" ), "Resources/afm/HV______.AFM" );
        afmResources.put( COSName.getPDFName( "Helvetica-Bold" ), "Resources/afm/HVB_____.AFM" );
        afmResources.put( COSName.getPDFName( "Helvetica-BoldOblique" ), "Resources/afm/HVBO____.AFM" );
        afmResources.put( COSName.getPDFName( "Helvetica-Oblique" ), "Resources/afm/HVO_____.AFM" );
        afmResources.put( COSName.getPDFName( "Anna" ), "Resources/afm/IARG____.AFM" );
        afmResources.put( COSName.getPDFName( "PopplLaudatio-Italic" ), "Resources/afm/PYI_____.AFM" );
        afmResources.put( COSName.getPDFName( "PopplLaudatio-Medium" ), "Resources/afm/PYM_____.AFM" );
        afmResources.put( COSName.getPDFName( "PopplLaudatio-MediumItalic" ), "Resources/afm/PYMI____.AFM" );
        afmResources.put( COSName.getPDFName( "PopplLaudatio-Regular" ), "Resources/afm/PYRG____.AFM" );
        afmResources.put( COSName.getPDFName( "Boulevard" ), "Resources/afm/QTRG____.AFM" );
        afmResources.put( COSName.getPDFName( "Symbol" ), "Resources/afm/SY______.AFM" );
        afmResources.put( COSName.getPDFName( "Times-Bold" ), "Resources/afm/TIB_____.AFM" );
        afmResources.put( COSName.getPDFName( "Times-BoldItalic" ), "Resources/afm/TIBI____.AFM" );
        afmResources.put( COSName.getPDFName( "Times-Italic" ), "Resources/afm/TII_____.AFM" );
        afmResources.put( COSName.getPDFName( "Times-Roman" ), "Resources/afm/TIR_____.AFM" );
        afmResources.put( COSName.getPDFName( "Boton-Italic" ), "Resources/afm/TNI_____.AFM" );
        afmResources.put( COSName.getPDFName( "Boton-Medium" ), "Resources/afm/TNM_____.AFM" );
        afmResources.put( COSName.getPDFName( "Boton-MediumItalic" ), "Resources/afm/TNMI____.AFM" );
        afmResources.put( COSName.getPDFName( "Boton-Regular" ), "Resources/afm/TNR_____.AFM" );
        afmResources.put( COSName.getPDFName( "BaskervilleBE-Italic" ), "Resources/afm/VII_____.AFM" );
        afmResources.put( COSName.getPDFName( "BaskervilleBE-Medium" ), "Resources/afm/VIM_____.AFM" );
        afmResources.put( COSName.getPDFName( "BaskervilleBE-MediumItalic" ), "Resources/afm/VIMI____.AFM" );
        afmResources.put( COSName.getPDFName( "BaskervilleBE-Regular" ), "Resources/afm/VIRG____.AFM" );
        afmResources.put( COSName.getPDFName( "Giddyup" ), "Resources/afm/WG______.AFM" );
        afmResources.put( COSName.getPDFName( "Giddyup-Thangs" ), "Resources/afm/WGTHA______.AFM" );
        afmResources.put( COSName.getPDFName( "ZapfDingbats" ), "Resources/afm/ZapfDingbats.AFM" );
    }

    /**
     * This will clear AFM resources that are stored statically.
     * This is usually not a problem unless you want to reclaim
     * resources for a long running process.
     *
     * SPECIAL NOTE: The font calculations are currently in COSObject, which
     * is where they will reside until PDFont is mature enough to take them over.
     * PDFont is the appropriate place for them and not in COSObject but we need font
     * calculations for text extractaion.  THIS METHOD WILL BE MOVED OR REMOVED
     * TO ANOTHER LOCATION IN A FUTURE VERSION OF PDFBOX.
     */
    public static void clearAFMResources()
    {
        afmObjects = new HashMap();
    }

    /**
     * Constructor
     */
    public PDFont()
    {
        font = new COSDictionary();
        font.setItem( COSName.TYPE, COSName.FONT );
    }

    /**
     * Constructor
     *
     * @param fontDictionary The font dictionary according to the PDF specification.
     */
    public PDFont( COSDictionary fontDictionary )
    {
        font = fontDictionary;
    }

    /**
     * @see COSObjectable#getCOSObject()
     */
    public COSBase getCOSObject()
    {
        return font;
    }

    /**
     * This will get the font width for a character.
     *
     * @param c The character code to get the width for.
     * @param offset The offset into the array.
     * @param length The length of the data.
     *
     * @return The width is in 1000 unit of text space, ie 333 or 777
     *
     * @throws IOException If an error occurs while parsing.
     */
    public float getFontWidth( byte[] c, int offset, int length ) throws IOException
    {
        float fontWidth = 0;
        int code = getCodeFromArray( c, offset, length );

        //hmm should this be in a subclass??
        COSInteger firstChar = (COSInteger)font.getDictionaryObject( COSName.FIRSTCHAR );
        COSInteger lastChar = (COSInteger)font.getDictionaryObject( COSName.LASTCHAR );
        if( firstChar != null && lastChar != null )
        {
            long first = firstChar.intValue();
            long last = lastChar.intValue();
            if( code >= first && code <= last && font.getDictionaryObject( COSName.WIDTHS ) != null )
            {
                COSArray widthArray = (COSArray)font.getDictionaryObject( COSName.WIDTHS );
                COSNumber fontWidthObject = (COSNumber)widthArray.get( (int)(code - first) );
                fontWidth = fontWidthObject.floatValue();
            }
            else
            {
                fontWidth = getFontWidthFromAFMFile( code );
            }
        }
        else
        {
            fontWidth = getFontWidthFromAFMFile( code );
        }
        return fontWidth;
    }

    /**
     * Used for multibyte encodings.
     *
     * @param data The array of data.
     * @param offset The offset into the array.
     * @param length The number of bytes to use.
     */
    private int getCodeFromArray( byte[] data, int offset, int length )
    {
        int code = 0;
        for( int i=0; i<length; i++ )
        {
            code <<= 8;
            code = (data[offset+i]+256)%256;
        }
        return code;
    }

    /**
     * This will attempt to get the font width from an AFM file.
     *
     * @param code The character code we are trying to get.
     *
     * @return The font width from the AFM file.
     *
     * @throws IOException if we cannot find the width.
     */
    private float getFontWidthFromAFMFile( int code ) throws IOException
    {
        float retval = 0;
        COSName baseFont = (COSName)font.getDictionaryObject( COSName.BASE_FONT );
        if( baseFont != null )
        {
            FontMetric metric = getAFM( baseFont );
            if( metric != null )
            {
                Encoding encoding = getEncoding();
                COSName characterName = encoding.getName( code );
                retval = metric.getCharacterWidth( characterName.getName() );
            }
        }
        return retval;
    }

    /**
     * This will get an AFM object.
     *
     * @param name The name of the AFM object.
     *
     * @return The afm object from the name.
     *
     * @throws IOException If there is an error getting the AFM object.
     */
    private FontMetric getAFM( COSName name ) throws IOException
    {
        FontMetric result = (FontMetric)afmObjects.get( name );
        if( result == null )
        {
            String resource = (String)afmResources.get( name );
            if( log.isDebugEnabled() )
            {
                log.debug("resource: " + resource + ", name: " + name.getName());
            }
            if( resource == null )
            {
                //ok for now
                //throw new IOException( "Unknown AFM font '" + name.getName() + "'" );
            }
            else
            {
                ClassLoader loader = getClass().getClassLoader();
                InputStream afmStream = loader.getResourceAsStream( resource );
                if( afmStream == null )
                {
                    throw new IOException( "Can't handle font width:" + resource );
                }
                AFMParser parser = new AFMParser( afmStream );
                parser.parse();
                result = parser.getResult();
                afmObjects.put( name, result );
            }
        }
        return result;
    }

    /**
     * This will perform the encoding of a character if needed.
     *
     * @param c The character to encode.
     * @param offset The offset into the array to get the data
     * @param length The number of bytes to read.
     *
     * @return The value of the encoded character.
     *
     * @throws IOException If there is an error during the encoding.
     */
    public String encode( byte[] c, int offset, int length ) throws IOException
    {
        String retval = new String( c );
        COSName fontSubtype = (COSName)font.getDictionaryObject( COSName.getPDFName( "Subtype" ) );
        if( (fontSubtype.getName().equals( "Type0" ) ||
            fontSubtype.getName().equals( "TrueType" ) ) &&
            font.getDictionaryObject( COSName.getPDFName( "ToUnicode" ) ) != null )
        //if( getDictionaryObject( COSName.getPDFName( "ToUnicode" ) ) != null )
        {
            retval = "";
            COSStream toUnicode = (COSStream)font.getDictionaryObject( COSName.getPDFName( "ToUnicode" ) );
            if( toUnicode != null )
            {
                if( cmap == null )
                {
                    CMapParser parser = new CMapParser( toUnicode.getUnfilteredStream(), toUnicode.getScratchFile() );
                    parser.parse();
                    cmap = parser.getResult();
                }
                retval = cmap.lookup( c, offset, length );
            }
        }
        else
        {
            Encoding encoding = getEncoding();
            if( encoding != null )
            {
                retval = encoding.getCharacter( getCodeFromArray( c, offset, length ) );
            }
        }

        /**
         //This is experimental and is not anywheres ready for release.
        if( fontSubtype.equals( COSName.getPDFName( "Type3" ) ) )
        {
            //we are actually an inlined stream, we need to do some OCR work
            //to find out the actual character that is being encoded
            org.pdfbox.ocr.FontImageDataConverter converter = new org.pdfbox.ocr.FontImageDataConverter();
            org.pdfbox.ocr.OCRImage image = converter.createImageForCharacter( this, retval );
            if( image != null )
            {
                org.pdfbox.ocr.OCR ocr = new org.pdfbox.ocr.OCR();
                String value = ocr.decodeImage( image );
            }
        }
        else
        {
            //we will assume that what we have is good enough
        }**/
        return retval;
    }

    /**
     * This will get or create the encoder.
     *
     * @return The encoding to use.
     *
     * @throws IOException If there is an error getting the encoding.
     */
    private Encoding getEncoding() throws IOException
    {
        EncodingManager manager = new EncodingManager();
        if( fontEncoding == null )
        {
            COSBase encoding = font.getDictionaryObject( COSName.ENCODING );
            if( encoding == null )
            {
                COSName baseFont = (COSName)font.getDictionaryObject( COSName.BASE_FONT );
                if( baseFont != null )
                {
                    if( afmResources.get( baseFont ) != null )
                    {
                        fontEncoding = new AFMEncoding( getAFM( baseFont ) );
                    }
                }
                if( fontEncoding == null )
                {
                    fontEncoding = manager.getStandardEncoding();
                }
            }
            else if( encoding instanceof COSDictionary )
            {
                COSDictionary encodingDic = (COSDictionary)font.getDictionaryObject( COSName.ENCODING );
                fontEncoding = new DictionaryEncoding( encodingDic );
            }
            else if( encoding instanceof COSName )
            {
                if( !encoding.equals( COSName.IDENTITY_H ) )
                {
                    fontEncoding = manager.getEncoding( (COSName)encoding );
                }
            }
            else
            {
                throw new IOException( "Unexpected encoding type:" + encoding.getClass().getName() );
            }
        }
        return fontEncoding;
    }
}