/**
 * Copyright (c) 2003, www.pdfbox.org
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 3. Neither the name of pdfbox; nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * http://www.pdfbox.org
 *
 */
package org.pdfbox.util;

/**
 * This class will be used for matrix manipulation.
 *
 * @author Ben Litchfield (ben@csh.rit.edu)
 * @version $Revision: 1.6 $
 */
public class Matrix
{
    private static final int MATRIX_SIZE = 3;
    private float[][] matrix = new float[MATRIX_SIZE][MATRIX_SIZE];

    /**
     * Constructor.
     */
    public Matrix()
    {
        matrix = new float[MATRIX_SIZE][MATRIX_SIZE];

        this.setValue( 0, 0, 1 );
        this.setValue( 0, 1, 0 );
        this.setValue( 0, 2, 0 );
        this.setValue( 1, 0, 0 );
        this.setValue( 1, 1, 1 );
        this.setValue( 1, 2, 0 );
        this.setValue( 2, 0, 0 );
        this.setValue( 2, 1, 0 );
        this.setValue( 2, 2, 1 );
    }

    /**
     * This is used for debugging purposes only.  It will print the contents
     * of the matrix to stdout.
     */
    public void print()
    {
        for( int i=0; i<MATRIX_SIZE; i++ )
        {
            System.out.print( "[" );
            for( int j=0; j<MATRIX_SIZE; j++ )
            {
                System.out.print( matrix[i][j] );
                if( j < MATRIX_SIZE-1 )
                {
                    System.out.print( "," );
                }
            }
            System.out.println( "]" );
        }
    }

    /**
     * This will get a matrix value at some point.
     *
     * @param row The row to get the value from.
     * @param column The column to get the value from.
     *
     * @return The value at the row/column position.
     */
    public float getValue( int row, int column )
    {
        return matrix[row][column];
    }

    /**
     * This will set a value at a position.
     *
     * @param row The row to set the value at.
     * @param column the column to set the value at.
     * @param value The value to set at the position.
     */
    public void setValue( int row, int column, float value )
    {
        matrix[row][column] = value;
    }

    /**
     * This will take the current matrix and multipy it with a matrix that is passed in.
     *
     * @param b The matrix to multiply by.
     *
     * @return The result of the two multiplied matrices.
     */
    public Matrix multiply( Matrix b )
    {
        Matrix result = new Matrix();

        float[][] bMatrix = b.matrix;
        float[][] resultMatrix = result.matrix;
        resultMatrix[0][0] = matrix[0][0] * bMatrix[0][0] + matrix[0][1] * bMatrix[1][0] + matrix[0][2] * bMatrix[2][0];
        resultMatrix[0][1] = matrix[0][0] * bMatrix[0][1] + matrix[0][1] * bMatrix[1][1] + matrix[0][2] * bMatrix[2][1];
        resultMatrix[0][2] = matrix[0][0] * bMatrix[0][2] + matrix[0][1] * bMatrix[1][2] + matrix[0][2] * bMatrix[2][2];
        resultMatrix[1][0] = matrix[1][0] * bMatrix[0][0] + matrix[1][1] * bMatrix[1][0] + matrix[1][2] * bMatrix[2][0];
        resultMatrix[1][1] = matrix[1][0] * bMatrix[0][1] + matrix[1][1] * bMatrix[1][1] + matrix[1][2] * bMatrix[2][1];
        resultMatrix[1][2] = matrix[1][0] * bMatrix[0][2] + matrix[1][1] * bMatrix[1][2] + matrix[1][2] * bMatrix[2][2];
        resultMatrix[2][0] = matrix[2][0] * bMatrix[0][0] + matrix[2][1] * bMatrix[1][0] + matrix[2][2] * bMatrix[2][0];
        resultMatrix[2][1] = matrix[2][0] * bMatrix[0][1] + matrix[2][1] * bMatrix[1][1] + matrix[2][2] * bMatrix[2][1];
        resultMatrix[2][2] = matrix[2][0] * bMatrix[0][2] + matrix[2][1] * bMatrix[1][2] + matrix[2][2] * bMatrix[2][2];

        return result;
    }

    /**
     * This will copy the text matrix data.
     *
     * @return a matrix that matches this one.
     */
    public Matrix copy()
    {
        Matrix copy = new Matrix();
        for( int i=0; i<MATRIX_SIZE; i++ )
        {
            for( int j=0; j<MATRIX_SIZE; j++ )
            {
                copy.matrix[i][j] = this.matrix[i][j];
            }
        }
        return copy;
    }

    /**
     * This will return a string representation of the matrix
     *
     * @return The matrix as a string.
     */
    public String toString()
    {
        StringBuffer result = new StringBuffer( "\n" );
        for( int i=0; i<MATRIX_SIZE; i++ )
        {
            result.append( "[" );
            for( int j=0; j<MATRIX_SIZE; j++ )
            {
                result.append( "" + matrix[i][j] + " " );
            }
            result.append( "]\n" );
        }
        return result.toString();
    }

}