/**
 * Copyright (c) 2003, www.pdfbox.org
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 3. Neither the name of pdfbox; nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * http://www.pdfbox.org
 *
 */
package org.pdfbox.util;

import java.io.IOException;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Stack;

import org.pdfbox.cos.COSArray;
import org.pdfbox.cos.COSBase;
import org.pdfbox.cos.COSNumber;
import org.pdfbox.cos.COSName;
import org.pdfbox.cos.COSObject;
import org.pdfbox.cos.COSStream;
import org.pdfbox.cos.COSString;

import org.pdfbox.pdmodel.font.PDFont;

import org.apache.log4j.Category;

/**
 * This class will run through a PDF content stream and execute certain operations
 * and provide a callback interface for clients that want to do things with the stream.
 * See the PDFTextStripper class for an example of how to use this class.
 *
 * @author Ben Litchfield (ben@csh.rit.edu)
 * @version $Revision: 1.5 $
 */
public class PDFStreamEngine
{
    private static Category log = Category.getInstance(PDFStreamEngine.class.getName());

    private COSStream stream;

    private Matrix textMatrix = null;
    private Matrix textLineMatrix = null;
    private Matrix currentTransformationMatrix = null;
    private Stack ctmStack = new Stack();
    private float currentLeading = 0;
    private float currentCharacterSpacing = 0;
    private float currentWordSpacing = 0;
    private PDFont currentFont;
    private COSNumber currentFontSize;
    private Map fonts = null;

    /**
     * Constructor
     *
     */
    public PDFStreamEngine()
    {
    }

    /**
     * This will process the contents of the stream.
     *
     * @param cosStream the Stream to execute.
     * @param fontMap The fonts for the current stream.
     *
     * @throws IOException if there is an error accessing the stream.
     */
    public void processStream( COSStream cosStream, Map fontMap ) throws IOException
    {
        stream = cosStream;
        textMatrix = null;
        textLineMatrix = null;
        currentTransformationMatrix = new Matrix();
        ctmStack = new Stack();
        currentLeading = 0;
        currentCharacterSpacing = 0;
        currentWordSpacing = 0;
        currentFont = null;
        currentFontSize = null;
        fonts = fontMap;

        List arguments = new ArrayList();
        long startTokens = System.currentTimeMillis();
        List tokens = stream.getStreamTokens();
        long stopTokens = System.currentTimeMillis();
        if( log.isDebugEnabled() )
        {
            log.debug( "Getting tokens time=" + (stopTokens-startTokens) );
        }
        if( tokens != null )
        {
            Iterator iter = tokens.iterator();
            while( iter.hasNext() )
            {
                Object next = iter.next();
                if( next instanceof COSObject )
                {
                    arguments.add( ((COSObject)next).getObject() );
                }
                else if( next instanceof PDFOperator )
                {
                    processOperator( (PDFOperator)next, arguments );
                    arguments = new ArrayList();
                }
                else
                {
                    arguments.add( next );
                }
            }
        }
    }

    /**
     * You should override this method if you want to perform an action when a
     * string is being shown.
     *
     * @param string The string to display.
     *
     * @return A description of the text being shown.
     *
     * @throws IOException If there is an error showing the string
     */
    protected TextPosition showString( byte[] string ) throws IOException
    {
        float xPos = 0;
        float yPos = 0;
        float width = 0;
        float fontSize = currentFontSize.floatValue();
        float lastCharWidth = fontSize; // This is a worst-case value
        float wordSpacing = 0;
        float padding = 0;
        StringBuffer encodedString = new StringBuffer( string.length );

        Matrix textRenderingMatrix = textMatrix.multiply(this.currentTransformationMatrix);

        //this
        int codeLength = 1;
        for( int i=0; i<string.length; i+=codeLength )
        {
            codeLength = 1;
            float characterWidth;
            /*
            log.debug( "Checking code '" + characterCode + "' font=" + currentFont +
                 " Tc=" + currentCharacterSpacing +
                 " Tw=" + currentWordSpacing +
                 " width=" + width );
             */
            String c = currentFont.encode( string, i, codeLength );
            if( c == null && i+1<string.length)
            {
                //maybe a multibyte encoding
                codeLength++;
                c = currentFont.encode( string, i, codeLength );
                characterWidth = (currentFont.getFontWidth( string, i, codeLength )/1000)*fontSize;
            }
            else
            {
                characterWidth = (currentFont.getFontWidth( string, i, codeLength )/1000)*fontSize;
            }
            if( c != null )
            {
                encodedString.append( c );
            }


            if( c != null )
            {
                // PDF Spec - 5.5.2 Word Spacing
                //
                // Word spacing works the same was as character spacing, but applies
                // only to the space character, code 32.
                //
                // Note: Word spacing is applied to every occurrence of the single-byte
                // character code 32 in a string.  This can occur when using a simple
                // font or a composite font that defines code 32 as a single-byte code.
                // It does not apply to occurrences of the byte value 32 in multiple-byte
                // codes.
                //
                // RDD - My interpretation of this is that only character code 32's that
                // encode to spaces should have word spacing applied.  Cases have been
                // observed where a font has a space character with a character code
                // other than 32, and where word spacing (Tw) was used.  In these cases,
                // applying word spacing to either the non-32 space or to the character
                // code 32 non-space resulted in errors consistent with this interpretation.
                //
                if ((string[i] == 0x20) && c.equals( " " ) )
                {
                    padding = currentWordSpacing;
                }
                else
                {
                    padding = currentCharacterSpacing;
                }
            }
            if( i == 0 )
            {
                xPos = textRenderingMatrix.getValue( 2, 0 );
                yPos = textRenderingMatrix.getValue( 2, 1 );

                // RDD - Attempt to compute a "wordSpacing" value that represents the space between
                // words for space determination by the TextPosition user.
                //
                byte[] baseWidthChar = new byte[]{ (byte)'n'};
                if ("n".equals( currentFont.encode(baseWidthChar, 0, 1) ) )
                {
                    wordSpacing = (currentFont.getFontWidth( baseWidthChar, 0, 1 )/1000)*fontSize;
                }

                if (wordSpacing == 0)
                {
                    // If we can't a space character, or if its width was zero, we
                    // use the width of the current character.
                    //
                    wordSpacing = characterWidth;
                }

                if (wordSpacing == 0)
                {
                    // If we can't get any character width at all, we approximate
                    // using a percentage of the font height.
                    //
                    wordSpacing = fontSize * 0.075f;
                }

                wordSpacing += currentWordSpacing;
            }

            if (characterWidth != 0)
            {
                // RDD - Track the last character width so that it can be used for
                // rough kerning approximation by the TextPosition user.
                //
                lastCharWidth = characterWidth;
            }

            width += characterWidth + padding;
        }

        // We want to update the textMatrix using the width, in text space units.
        //
        Matrix td = new Matrix();
        td.setValue( 2, 0, width * textMatrix.getValue(0,0));
        //log.debug( "TextMatrix before " + textMatrix );
        textMatrix = textMatrix.multiply( td );
        //log.debug( "TextMatrix after " + textMatrix );

        // Scale everything to convert to textRendering (device) units.
        //
        float xScale = textRenderingMatrix.getValue(0,0);
        float yScale = textRenderingMatrix.getValue(1,1);

        width *= xScale;
        fontSize *= yScale;
        lastCharWidth *= xScale;
        wordSpacing *= xScale;
        padding *= xScale;

        if (log.isDebugEnabled())
        {
            log.debug("TextPosition" +
                      " x=" + xPos +
                      " y=" + yPos +
                      " width=" +  width +
                      " string=\"" + encodedString.toString() + "\"" +
                      " fontSize=" + fontSize +
                      " lastCharWidth=" + lastCharWidth +
                      " wordSpacing=" + wordSpacing +
                      " padding=" + padding);
        }

        TextPosition position = new TextPosition( xPos, yPos, width, encodedString.toString(),
                                                  fontSize, lastCharWidth, wordSpacing, padding );
        return position;
    }

    /**
     * This is used to handle an operation.
     *
     * @param operator The operation to perform.
     * @param arguments The list of arguments.
     *
     * @throws IOException If there is an error processing the operation.
     */
    protected void processOperator( PDFOperator operator, List arguments ) throws IOException
    {
        String operation = operator.getOperation();
        if( log.isDebugEnabled() )
        {
            log.debug( "processOperator( '" + operation + "' )" );
        }
        /**
        if( operation.equals( "b" ) )
        {
            //Close, fill, and stroke path using nonzero winding number rule
        }
        else if( operation.equals( "B" ) )
        {
            //Fill and stroke path using nonzero winding number rule
        }
        else if( operation.equals( "b*" ) )
        {
            //Close, fill, and stroke path using even-odd rule
        }
        else if( operation.equals( "B*" ) )
        {
            //Fill and stroke path using even-odd rule
        }
        else if( operation.equals( "BDC" ) )
        {
            //(PDF 1.2) Begin marked-content sequence with property list
        }
        else if( operation.equals( "BI" ) )
        {
            //begin inline image object
        }
        else if( operation.equals( "BMC" ) )
        {
            //(PDF 1.2) Begin marked-content sequence
        }
        else */if( operation.equals( "BT" ) )
        {
            log.debug( "<BT>" );
            textMatrix = new Matrix();
            textLineMatrix = textMatrix.copy();
        }/*
        else if( operation.equals( "BX" ) )
        {
            //(PDF 1.1) Begin compatibility section
        }
        else if( operation.equals( "c" ) )
        {
            //Append curved segment to path (three control points)
        }*/
        else if( operation.equals( "cm" ) )
        {
            //concatenate matrix to current transformation matrix
            COSNumber a = (COSNumber)arguments.get( 0 );
            COSNumber b = (COSNumber)arguments.get( 1 );
            COSNumber c = (COSNumber)arguments.get( 2 );
            COSNumber d = (COSNumber)arguments.get( 3 );
            COSNumber e = (COSNumber)arguments.get( 4 );
            COSNumber f = (COSNumber)arguments.get( 5 );

            if (log.isDebugEnabled())
            {
                log.debug("<cm " +
                          "a=\"" + a.floatValue() + "\" " +
                          "b=\"" + b.floatValue() + "\" " +
                          "c=\"" + c.floatValue() + "\" " +
                          "d=\"" + d.floatValue() + "\" " +
                          "e=\"" + e.floatValue() + "\" " +
                          "f=\"" + f.floatValue() + "\" >");
            }

            Matrix newMatrix = new Matrix();
            newMatrix.setValue( 0, 0, a.floatValue() );
            newMatrix.setValue( 0, 1, b.floatValue() );
            newMatrix.setValue( 1, 0, c.floatValue() );
            newMatrix.setValue( 1, 1, d.floatValue() );
            newMatrix.setValue( 2, 0, e.floatValue() );
            newMatrix.setValue( 2, 1, f.floatValue() );

            currentTransformationMatrix = currentTransformationMatrix.multiply(newMatrix);
        }/*
        else if( operation.equals( "cs" ) )
        {
            //set colorspace for non stroking operations
            COSName name = (COSName)parameterStack.get( parameterStack.size() -2 );
            //System.out.println( "<cs name=\"" + name.getName() + "\" >" );
        }
        else if( operation.equals( "CS" ) )
        {
            //(PDF 1.1) Set color space for stroking operations
            COSName name = (COSName)parameterStack.get( parameterStack.size() -2 );
            //System.out.println( "<CS name=\"" + name.getName() + "\" >" );
        }
        else if( operation.equals( "d" ) )
        {
            //Set the line dash pattern in the graphics state
            //System.out.println( "<d>" );
        }
        else if( operation.equals( "d0" ) )
        {
            COSNumber x = (COSNumber)parameterStack.get( parameterStack.size() -3 );
            COSNumber y = (COSNumber)parameterStack.get( parameterStack.size() -2 );
            //System.out.println( "<d0 x=\"" + x.getValue() + "\" y=\"" + y.getValue() + "\">" );
        }
        else if( operation.equals( "d1" ) )
        {
            //System.out.println( "<d1>" );
        }
        else if( operation.equals( "Do" ) )
        {
            //invoke named object.
        }
        else if( operation.equals( "DP" ) )
        {
            //(PDF 1.2) De.ne marked-content point with property list
        }
        else if( operation.equals( "EI" ) )
        {
            //end inline image object
        }
        else if( operation.equals( "EMC" ) )
        {
            //End inline image object
        }*/
        else if( operation.equals( "ET" ) )
        {
            log.debug( "<ET>" );
            textMatrix = null;
            textLineMatrix = null;
        }/*
        else if( operation.equals( "EX" ) )
        {
            //(PDF 1.1) End compatibility section
        }
        else if( operation.equals( "f" ) )
        {
            //Fill the path, using the nonzero winding number rule to determine the region to .ll
            //System.out.println( "<f>" );
        }
        else if( operation.equals( "F" ) )
        {
            //Equivalent to f; included only for compatibility. Although applications that read
            //PDF files must be able to accept this operator, those that generate PDF files should
            //use f instead.
            //System.out.println( "<F>" );
        }
        else if( operation.equals( "f*" ) )
        {
            //Fill path using even-odd rule
        }
        else if( operation.equals( "g" ) )
        {
            //set gray level for nonstroking operations
            COSNumber grayLevel = (COSNumber)parameterStack.get( parameterStack.size() -2 );
            //System.out.println( "<g gray=\"" + grayLevel.getValue() + "\" />" );
        }
        else if( operation.equals( "G" ) )
        {
            //set gray level for stroking operations
            COSNumber grayLevel = (COSNumber)parameterStack.get( parameterStack.size() -2 );
            //System.out.println( "<G gray=\"" + grayLevel.getValue() + "\" />" );
        }
        else if( operation.equals( "gs" ) )
        {
            //set parameters from graphics state parameter dictionary
        }
        else if( operation.equals( "h" ) )
        {
            //close subpath
        }
        else if( operation.equals( "i" ) )
        {
            //set flatness tolerance, not sure what this does
        }
        else if( operation.equals( "ID" ) )
        {
            //begin inline image data
        }
        else if( operation.equals( "j" ) )
        {
            //Set the line join style in the graphics state
            //System.out.println( "<j>" );
        }
        else if( operation.equals( "J" ) )
        {
            //Set the line cap style in the graphics state
            //System.out.println( "<J>" );
        }
        else if( operation.equals( "k" ) )
        {
            //Set CMYK color for nonstroking operations
        }
        else if( operation.equals( "K" ) )
        {
            //Set CMYK color for stroking operations
        }
        else if( operation.equals( "l" ) )
        {
            //append straight line segment from the current point to the point.
            COSNumber x = (COSNumber)parameterStack.get( parameterStack.size() -3 );
            COSNumber y = (COSNumber)parameterStack.get( parameterStack.size() -2 );
            //System.out.println( "<l x=\"" + x.getValue() + "\" y=\"" + y.getValue() + "\" >" );
        }
        else if( operation.equals( "m" ) )
        {
            COSNumber x = (COSNumber)parameterStack.get( parameterStack.size() -3 );
            COSNumber y = (COSNumber)parameterStack.get( parameterStack.size() -2 );
            //System.out.println( "<m x=\"" + x.getValue() + "\" y=\"" + y.getValue() + "\" >" );
        }
        else if( operation.equals( "M" ) )
        {
            //System.out.println( "<M>" );
        }
        else if( operation.equals( "MP" ) )
        {
            //(PDF 1.2) Define marked-content point
        }
        else if( operation.equals( "n" ) )
        {
            //End path without .lling or stroking
            //System.out.println( "<n>" );
        }*/
        else if( operation.equals( "q" ) )
        {
            //save graphics state
            if( log.isDebugEnabled() )
            {
                log.debug( "<" + operation + "> - save state" );
            }
            ctmStack.push(currentTransformationMatrix);
        }
        else if( operation.equals( "Q" ) )
        {
            //restore graphics state
            if( log.isDebugEnabled() )
            {
                log.debug( "<" + operation + "> - restore state" );
            }
            currentTransformationMatrix = (Matrix)ctmStack.pop();
        }/*
        else if( operation.equals( "re" ) )
        {
            COSNumber x = (COSNumber)parameterStack.get( parameterStack.size() -5 );
            COSNumber y = (COSNumber)parameterStack.get( parameterStack.size() -4 );
            COSNumber width = (COSNumber)parameterStack.get( parameterStack.size() -3 );
            COSNumber height = (COSNumber)parameterStack.get( parameterStack.size() -2 );
            //System.out.println( "<re x=\"" + x.getValue() + "\" y=\"" + y.getValue() + "\" width=\"" +
            //                                 width.getValue() + "\" height=\"" + height.getValue() + "\" >" );
        }
        else if( operation.equals( "rg" ) )
        {
            //Set RGB color for nonstroking operations
        }
        else if( operation.equals( "RG" ) )
        {
            //Set RGB color for stroking operations
        }
        else if( operation.equals( "ri" ) )
        {
            //Set color rendering intent
        }
        else if( operation.equals( "s" ) )
        {
            //Close and stroke path
        }
        else if( operation.equals( "S" ) )
        {
            //System.out.println( "<S>" );
        }
        else if( operation.equals( "sc" ) )
        {
            //set color for nonstroking operations
            //System.out.println( "<sc>" );
        }
        else if( operation.equals( "SC" ) )
        {
            //set color for stroking operations
            //System.out.println( "<SC>" );
        }
        else if( operation.equals( "scn" ) )
        {
            //set color for nonstroking operations special
        }
        else if( operation.equals( "SCN" ) )
        {
            //set color for stroking operations special
        }
        else if( operation.equals( "sh" ) )
        {
            //(PDF 1.3) Paint area de.ned by shading pattern
        }*/
        else if( operation.equals( "T*" ) )
        {
            if (log.isDebugEnabled())
            {
                log.debug("<T* currentLeading=\"" + currentLeading + "\">");
            }
            //move to start of next text line
            if( currentLeading == 0 )
            {
                currentLeading = -.01f;
            }
            Matrix td = new Matrix();
            td.setValue( 2, 1, -1 * currentLeading * textMatrix.getValue(1,1));
            textLineMatrix = textLineMatrix.multiply( td );
            textMatrix = textLineMatrix.copy();
        }
        else if( operation.equals( "Tc" ) )
        {
            //set character spacing
            COSNumber characterSpacing = (COSNumber)arguments.get( 0 );
            if (log.isDebugEnabled())
            {
                log.debug("<Tc characterSpacing=\"" + characterSpacing.floatValue() + "\" />");
            }
            currentCharacterSpacing = characterSpacing.floatValue();
        }
        else if( operation.equals( "Td" ) )
        {
            COSNumber x = (COSNumber)arguments.get( 0 );
            COSNumber y = (COSNumber)arguments.get( 1 );
            if (log.isDebugEnabled())
            {
                log.debug("<Td x=\"" + x.floatValue() + "\" y=\"" + y.floatValue() + "\">");
            }
            Matrix td = new Matrix();
            td.setValue( 2, 0, x.floatValue() * textMatrix.getValue(0,0) );
            td.setValue( 2, 1, y.floatValue() * textMatrix.getValue(1,1) );
            //log.debug( "textLineMatrix before " + textLineMatrix );
            textLineMatrix = textLineMatrix.multiply( td );
            //log.debug( "textLineMatrix after " + textLineMatrix );
            textMatrix = textLineMatrix.copy();
        }
        else if( operation.equals( "TD" ) )
        {
            //move text position and set leading
            COSNumber x = (COSNumber)arguments.get( 0 );
            COSNumber y = (COSNumber)arguments.get( 1 );
            if (log.isDebugEnabled())
            {
                log.debug("<TD x=\"" + x.floatValue() + "\" y=\"" + y.floatValue() + "\">");
            }
            currentLeading = -1 * y.floatValue();
            Matrix td = new Matrix();
            td.setValue( 2, 0, x.floatValue() * textMatrix.getValue(0,0) );
            td.setValue( 2, 1, y.floatValue() * textMatrix.getValue(1,1) );
            //log.debug( "textLineMatrix before " + textLineMatrix );
            textLineMatrix = textLineMatrix.multiply( td );
            //log.debug( "textLineMatrix after " + textLineMatrix );
            textMatrix = textLineMatrix.copy();
        }
        else if( operation.equals( "Tf" ) )
        {
            //set font and size
            COSName fontName = (COSName)arguments.get( 0 );
            currentFontSize = (COSNumber)arguments.get( 1 );

            if (log.isDebugEnabled())
            {
                log.debug("<Tf font=\"" + fontName.getName() + "\" size=\"" + currentFontSize.floatValue() + "\">");
            }

            //old way
            //currentFont = (COSObject)stream.getDictionaryObject( fontName );
            //if( currentFont == null )
            //{
            //    currentFont = (COSObject)currentFontDictionary.getItem( fontName );
            //}
            currentFont = (PDFont)fonts.get( fontName );
            if( currentFont == null )
            {
                throw new IOException( "Current font is null" );
            }
            //log.debug( "Font Resource=" + fontResource );
            //log.debug( "Current Font=" + currentFont );
            //log.debug( "currentFontSize=" + currentFontSize );
        }
        else if( operation.equals( "Tj" ) )
        {
            COSString string = (COSString)arguments.get( 0 );
            TextPosition pos = showString( string.getBytes() );
            if (log.isDebugEnabled())
            {
                log.debug("<Tj string=\"" + string.getString() + "\">");
            }
        }
        else if( operation.equals( "TJ" ) )
        {
            Matrix td = new Matrix();

            COSArray array = (COSArray)arguments.get( 0 );
            for( int i=0; i<array.size(); i++ )
            {
                COSBase next = array.get( i );
                if( next instanceof COSNumber )
                {
                    float value = -1*
                                  (((COSNumber)next).floatValue()/1000) *
                                  currentFontSize.floatValue() *
                                  textMatrix.getValue(1,1);

                    if (log.isDebugEnabled())
                    {
                        log.debug( "<TJ(" + i + ") value=\"" + value +
                                   "\", param=\"" + ((COSNumber)next).floatValue() +
                                   "\", fontsize=\"" + currentFontSize.floatValue() + "\">" );
                    }
                    td.setValue( 2, 0, value );
                    textMatrix = textMatrix.multiply( td );
                }
                else if( next instanceof COSString )
                {
                    TextPosition pos = showString( ((COSString)next).getBytes() );
                    if (log.isDebugEnabled())
                    {
                        log.debug("<TJ(" + i + ") string=\"" + pos.getString() + "\">");
                    }
                }
                else
                {
                    throw new IOException( "Unknown type in array for TJ operation:" + next );
                }
            }
        }
        else if( operation.equals( "TL" ) )
        {
            COSNumber leading = (COSNumber)arguments.get( 0 );
            currentLeading = leading.floatValue();
            if (log.isDebugEnabled())
            {
                log.debug("<TL leading=\"" + leading.floatValue() + "\" >");
            }
        }
        else if( operation.equals( "Tm" ) )
        {
            //Set text matrix and text line matrix
            COSNumber a = (COSNumber)arguments.get( 0 );
            COSNumber b = (COSNumber)arguments.get( 1 );
            COSNumber c = (COSNumber)arguments.get( 2 );
            COSNumber d = (COSNumber)arguments.get( 3 );
            COSNumber e = (COSNumber)arguments.get( 4 );
            COSNumber f = (COSNumber)arguments.get( 5 );

            if (log.isDebugEnabled())
            {
                log.debug("<Tm " +
                          "a=\"" + a.floatValue() + "\" " +
                          "b=\"" + b.floatValue() + "\" " +
                          "c=\"" + c.floatValue() + "\" " +
                          "d=\"" + d.floatValue() + "\" " +
                          "e=\"" + e.floatValue() + "\" " +
                          "f=\"" + f.floatValue() + "\" >");
            }

            textMatrix = new Matrix();
            textMatrix.setValue( 0, 0, a.floatValue() );
            textMatrix.setValue( 0, 1, b.floatValue() );
            textMatrix.setValue( 1, 0, c.floatValue() );
            textMatrix.setValue( 1, 1, d.floatValue() );
            textMatrix.setValue( 2, 0, e.floatValue() );
            textMatrix.setValue( 2, 1, f.floatValue() );
            textLineMatrix = textMatrix.copy();
        }/**
        else if( operation.equals( "Tr" ) )
        {
            //Set text rendering mode
            //System.out.println( "<Tr>" );
        }
        else if( operation.equals( "Ts" ) )
        {
            //Set text rise
            //System.out.println( "<Ts>" );
        }**/
        else if( operation.equals( "Tw" ) )
        {
            //set word spacing
            COSNumber wordSpacing = (COSNumber)arguments.get( 0 );
            if (log.isDebugEnabled())
            {
                log.debug("<Tw wordSpacing=\"" + wordSpacing.floatValue() + "\" />");
            }
            currentWordSpacing = wordSpacing.floatValue();
        }/**
        else if( operation.equals( "Tz" ) )
        {
            //Set horizontal text scaling
        }
        else if( operation.equals( "v" ) )
        {
            //Append curved segment to path (initial point replicated)
        }
        else if( operation.equals( "w" ) )
        {
            //Set the line width in the graphics state
            //System.out.println( "<w>" );
        }
        else if( operation.equals( "W" ) )
        {
            //Set clipping path using nonzero winding number rule
            //System.out.println( "<W>" );
        }
        else if( operation.equals( "W*" ) )
        {
            //Set clipping path using even-odd rule
        }
        else if( operation.equals( "y" ) )
        {
            //Append curved segment to path (final point replicated)
        }*/
        else if( operation.equals( "'" ) )
        {
            // Move to start of next text line, and show text
            //
            COSString string = (COSString)arguments.get( 0 );
            if (log.isDebugEnabled())
            {
                log.debug("<' string=\"" + string.getString() + "\">");
            }

            Matrix td = new Matrix();
            td.setValue( 2, 1, -1 * currentLeading * textMatrix.getValue(1,1));
            textLineMatrix = textLineMatrix.multiply( td );
            textMatrix = textLineMatrix.copy();

            showString( string.getBytes() );
        }
        else if( operation.equals( "\"" ) )
        {
            //Set word and character spacing, move to next line, and show text
            //
            COSNumber wordSpacing = (COSNumber)arguments.get( 0 );
            COSNumber characterSpacing = (COSNumber)arguments.get( 1 );
            COSString string = (COSString)arguments.get( 2 );

            if (log.isDebugEnabled())
            {
                log.debug("<\" wordSpacing=\"" + wordSpacing +
                          "\", characterSpacing=\"" + characterSpacing +
                          "\", string=\"" + string.getString() + "\">");
            }

            currentCharacterSpacing = characterSpacing.floatValue();
            currentWordSpacing = wordSpacing.floatValue();

            Matrix td = new Matrix();
            td.setValue( 2, 1, -1 * currentLeading * textMatrix.getValue(1,1));
            textLineMatrix = textLineMatrix.multiply( td );
            textMatrix = textLineMatrix.copy();

            showString( string.getBytes() );
        }
    }
}