/**
 * Copyright (c) 2003, www.pdfbox.org
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 3. Neither the name of pdfbox; nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * http://www.pdfbox.org
 *
 */
package org.pdfbox.util;

/**
 * This represents a character and a position on the screen of those characters.
 *
 * @author Ben Litchfield (ben@csh.rit.edu)
 * @version $Revision: 1.5 $
 */
public class TextPosition
{
    private float x;
    private float y;
    private float width;
    private String c;
    private float fontSize;
    private float lastCharWidth;
    private float wordSpacing;
    private float padding;

    /**
     * Constructor.
     *
     * @param x The x coordinate of the character.
     * @param y The y coordinate of the character.
     * @param width The width of the character.
     * @param c The character to be displayed.
     * @param fontSize The new font size.
     * @param lastCharWidth The width of the last character.
     * @param wordSpacing The expected spacing between words.
     * @param padding The padding included in the width (any padding after the last character).
     */
    public TextPosition(
        float x,
        float y,
        float width,
        String c,
        float fontSize,
        float lastCharWidth,
        float wordSpacing,
        float padding
        )
    {
        this.x = x;
        this.y = y;
        this.width = width;
        this.c = c;
        this.fontSize = fontSize;
        this.lastCharWidth = lastCharWidth;
        this.wordSpacing = wordSpacing;
        this.padding = padding;
    }

    /**
     * This will the character that will be displayed on the screen.
     *
     * @return The character on the screen.
     */
    public String getString()
    {
        return c;
    }

    /**
     * This will get the x position of the character.
     *
     * @return The x coordinate of the character.
     */
    public float getX()
    {
        return x;
    }

    /**
     * This will get the y position of the character.
     *
     * @return The y coordinate of the character.
     */
    public float getY()
    {
        return y;
    }

    /**
     * This will get with width of this character.
     *
     * @return The width of this character.
     */
    public float getWidth()
    {
        return width;
    }

    /**
     * This will get the font size that this object is
     * suppose to be drawn at.
     *
     * @return The font size.
     */
    public float getFontSize()
    {
        return fontSize;
    }

    /**
     * This will get the width of the last character in the object.
     *
     * @return The width of the last character in the object.
     */
    public float getLastCharWidth()
    {
        return lastCharWidth;
    }

    /**
     * This will get the expected spacing between words.
     *
     * @return The expected spacing between words.
     */
    public float getWordSpacing()
    {
        return wordSpacing;
    }

    /**
     * This will get the padding included in the width, which is to say any
     * padding added after the last character.
     *
     * @return The padding applied to the width.
     */
    public float getPadding()
    {
        return padding;
    }
}
